# --
# Kernel/System/TicketSearchSimple.pm - быстрый простой поиск
# Copyright (C) 2013 Меркушов Викто merkushov.victor@gmail.com
# --
# This software comes with ABSOLUTELY NO WARRANTY. For details, see
# the enclosed file COPYING for license information (AGPL). If you
# did not receive this file, see http://www.gnu.org/licenses/agpl.txt.
# --

package Kernel::System::TicketSearchSimple;

use strict;
use warnings;

use vars qw(@ISA $VERSION);
$VERSION = qw($Revision: 1.0 $) [1];

=head1 NAME

Kernel::System::TicketSearchSimple - быстрый поиск по тикетам

=head1 SYNOPSIS

Упрощённый поиск по тикетам. Без сложный джойнов (только по первичным ключам) и группировок. 
Позволяет получить сразу данные по тикетам.

=over 4

=cut

=item TicketSearchSimple()

To find tickets in your system.

    my @TicketIDs = $TicketObject->TicketSearch(
        # result (required)
        Result => 'ARRAY' || 'HASH' || 'COUNT',

        # result limit
        Limit => 100,

        # ticket number (optional) as STRING or as ARRAYREF
        TicketNumber => '%123546%',
        TicketNumber => ['%123546%', '%123666%'],

        # ticket title (optional) as STRING or as ARRAYREF
        Title => '%SomeText%',
        Title => ['%SomeTest1%', '%SomeTest2%'],

        QueueIDs => [1, 42, 512],

        # use also sub queues of Queue|Queues in search
        UseSubQueues => 0,

        # You can use types like normal, ...
        TypeIDs => [3, 4],

        # You can use states like new, open, pending reminder, ...
        StateIDs => [3, 4],

        # You also can use real state types like new, open, closed,
        # pending reminder, pending auto, removed and merged.
        StateTypeIDs => [1, 2, 3],

        PriorityIDs => [1, 2, 3],
        ServiceIDs => [1, 2, 3],
        SLAIDs => [1, 2, 3],
        LockIDs => [1, 2, 3],
        OwnerIDs => [1, 12, 455, 32],
        ResponsibleIDs => [1, 12, 455, 32],

        WatchUserIDs => [ '12' ],

        # search in archive (optional)
        # if archiving is on, if not specified the search processes unarchived only
        # 'y' searches archived tickets, 'n' searches unarchived tickets
        # if specified together all tickets are searched
        ArchiveFlags => ['y', 'n'],

        # OrderBy and SortBy (optional)
        OrderBy => 'Down',  # Down|Up
        SortBy  => 'Age',   # Owner|Responsible|CustomerID|State|TicketNumber|Queue|Priority|Age|Type|Lock
                            # Changed|Title|Service|SLA|PendingTime|EscalationTime
                            # EscalationUpdateTime|EscalationResponseTime|EscalationSolutionTime

        # OrderBy and SortBy as ARRAY for sub sorting (optional)
        OrderBy => ['Down', 'Up'],
        SortBy  => ['Priority', 'Age'],

        # user search (UserID is required)
        UserID     => 123,
        Permission => 'ro' || 'rw',

        # CacheTTL, cache search result in seconds (optional)
        CacheTTL => 60 * 15,

        # Optional, default 0. To include the dynamic field values for this ticket on the return structure.
        DynamicFields => 0,

        # Optional, default 0.
        CustomerName => 0,
    );

Returns:

Result: 'ARRAY'

    @TicketIDs = ( { ticket_info }, {...} );

Result: 'HASH'

    %TicketIDs = (
        1 => { ticket_info },
        2 => { ... },
        3 => { ... },
    );

Result: 'COUNT'

    $TicketIDs = 123;

=cut

sub TicketSearchSimple {
    my ( $Self, %Param ) = @_;

    my $Result  = $Param{Result}  || 'HASH';
    my $OrderBy = $Param{OrderBy} || 'Down';
    my $SortBy  = $Param{SortBy}  || 'ID';
    my $Limit   = $Param{Limit}   || 1000;

    my %SortOptions = (
        ID                     => 'st.id',
        Lock                   => 'st.ticket_lock_id',
        TicketNumber           => 'st.tn',
        Title                  => 'st.title',
        Queue                  => 'q.name',
        Priority               => 'st.ticket_priority_id',
        Age                    => 'st.create_time_unix',
        Changed                => 'st.change_time',
        PendingTime            => 'st.until_time',
        TicketEscalation       => 'st.escalation_time',
        EscalationTime         => 'st.escalation_time',
        EscalationUpdateTime   => 'st.escalation_update_time',
        EscalationResponseTime => 'st.escalation_response_time',
        EscalationSolutionTime => 'st.escalation_solution_time',
    );

    my @FieldOptions = (
        { Key => 'TicketID',        Field => 'st.id' },
        { Key => 'TicketNumber',    Field => 'st.tn' },
        { Key => 'Title',           Field => 'st.title' },
        { Key => 'State',           Field => 'ts.name' },
        { Key => 'StateID',         Field => 'ts.id' },
        { Key => 'StateType',       Field => 'tst.name' },
        { Key => 'Priority',        Field => 'tp.name' },
        { Key => 'PriorityID',      Field => 'tp.id' },
        { Key => 'Lock',            Field => 'tlt.name' },
        { Key => 'LockID',          Field => 'st.ticket_lock_id' },
        { Key => 'Queue',           Field => 'q.name' },
        { Key => 'QueueID',         Field => 'st.queue_id' },
        { Key => 'CustomerID',      Field => 'st.customer_id' },
        { Key => 'CustomerUserID',  Field => 'st.customer_user_id' },
        { Key => 'Owner',           Field => 'u.login' },
        { Key => 'OwnerName',       Field => 'u.first_name' },
        { Key => 'OwnerSurname',    Field => 'u.last_name' },
        { Key => 'OwnerID',         Field => 'st.user_id' },
        { Key => 'Type',            Field => 'tt.name' },
        { Key => 'TypeID',          Field => 'st.type_id' },
        { Key => 'SLA',             Field => 'sla.name' },
        { Key => 'SLAID',           Field => 'st.sla_id' },
        { Key => 'Service',         Field => 's.name' },
        { Key => 'ServiceID',       Field => 'st.service_id' },
        { Key => 'Responsible',     Field => 'ur.login' },
        { Key => 'ResponsibleID',   Field => 'st.responsible_user_id' },
        { Key => 'Created',         Field => 'st.create_time' },
        { Key => 'CreateTimeUnix',  Field => 'st.create_time_unix' },
        { Key => 'CreateBy',        Field => 'st.create_by' },
        { Key => 'Changed',         Field => 'st.change_time' },
        { Key => 'ChangeBy',        Field => 'st.change_by' },
        { Key => 'ArchiveFlag',     Field => 'st.archive_flag' },
        { Key => 'Readed',          Field => 'tf.ticket_value' },
    );

    # check required params
    if ( !$Param{UserID} && !$Param{CustomerUserID} ) {
        $Self->{LogObject}->Log(
            Priority => 'error',
            Message  => 'Need UserID or CustomerUserID params for permission check!',
        );
        return;
    }

    # check types of given arguments
    for my $Key (
        qw(
        TypeIDs StateIDs StateTypeIDs
        LockIDs OwnerIDs ResponsibleIDs CreatedUserIDs QueueIDs
        PriorityIDs ServiceIDs SLAIDs
        )
        )
    {
        next if !$Param{$Key};
        next if ref $Param{$Key} eq 'ARRAY' && @{ $Param{$Key} };

        # log error
        $Self->{LogObject}->Log(
            Priority => 'error',
            Message  => "The given param '$Key' is invalid or an empty array reference!",
        );
        return;
    }

    # quote id array elements
    for my $Key (
        qw(
        TypeIDs StateIDs StateTypeIDs LockIDs OwnerIDs ResponsibleIDs
        QueueIDs PriorityIDs ServiceIDs SLAIDs
        )
        )
    {
        next if !$Param{$Key};

        # quote elements
        for my $Element ( @{ $Param{$Key} } ) {
            if ( !defined $Self->{DBObject}->Quote( $Element, 'Integer' ) ) {

                # log error
                $Self->{LogObject}->Log(
                    Priority => 'error',
                    Message  => "The given param '$Element' in '$Key' is invalid!",
                );
                return;
            }
        }
    }


    # Only fetch DynamicField data if a field was requested for searching or sorting
    my $ParamCheckString = ( join '', keys %Param ) || '';

    if ( ref $Param{SortBy} eq 'ARRAY' ) {
        $ParamCheckString .= ( join '', @{ $Param{SortBy} } );
    }
    elsif ( ref $Param{SortBy} ne 'HASH' ) {
        $ParamCheckString .= $Param{SortBy} || '';
    }

    # check sort/order by options
    my @SortByArray;
    my @OrderByArray;
    if ( ref $SortBy eq 'ARRAY' ) {
        @SortByArray  = @{$SortBy};
        @OrderByArray = @{$OrderBy};
    }
    else {
        @SortByArray  = ($SortBy);
        @OrderByArray = ($OrderBy);
    }

    for my $Count ( 0 .. $#SortByArray ) {
        if ( !$SortOptions{ $SortByArray[$Count] } ) {
            $Self->{LogObject}->Log(
                Priority => 'error',
                Message  => 'Need valid SortBy (' . $SortByArray[$Count] . ')!',
            );
            return;
        }
        if ( $OrderByArray[$Count] ne 'Down' && $OrderByArray[$Count] ne 'Up' ) {
            $Self->{LogObject}->Log(
                Priority => 'error',
                Message  => 'Need valid OrderBy (' . $OrderByArray[$Count] . ')!',
            );
            return;
        }
    }

    if ( $Param{CustomerName} ) {
        push @FieldOptions, (
            { Key => 'CustomerName',      Field => 'cu.first_name' },
            { Key => 'CustomerSurname',   Field => 'cu.last_name' },
        );
    }

    # create sql
    my $SQLSelect;
    if ( $Result eq 'COUNT' ) {
        $SQLSelect = 'SELECT COUNT(DISTINCT(st.id))';
    }
    else {
        my $FieldsString = join ( ', ', map { $_->{Field} } @FieldOptions );
        $SQLSelect = 'SELECT ' . $FieldsString;
    }

    my $SQLFrom = ' 
        FROM ticket st 
            LEFT JOIN ticket_priority tp ON tp.id = st.ticket_priority_id
            LEFT JOIN ticket_lock_type tlt ON tlt.id = st.ticket_lock_id
            LEFT JOIN queue q ON q.id = st.queue_id 
            LEFT JOIN service s ON s.id = st.service_id
            LEFT JOIN sla ON sla.id = st.sla_id
            LEFT JOIN ticket_type tt ON tt.id = st.type_id 
            LEFT JOIN ticket_state ts ON ts.id = st.ticket_state_id 
            LEFT JOIN ticket_state_type tst ON tst.id = ts.type_id 
            LEFT JOIN users u ON u.id = st.user_id
            LEFT JOIN ticket_flag tf ON tf.ticket_id=st.id AND tf.create_by=u.id
            LEFT JOIN users ur ON ur.id = st.responsible_user_id
    ';

    my $SQLExt = ' WHERE tf.ticket_key=\'Seen\'';

    if ( $Param{TypeIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.type_id',
            IDRef       => $Param{TypeIDs},
        );
    }

    if ( $Param{StateIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.ticket_state_id',
            IDRef       => $Param{StateIDs},
        );
    }

    if ( $Param{StateTypeIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'tst.id',
            IDRef       => $Param{StateTypeIDs},
        );
    }

    if ( $Param{LockIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.ticket_lock_id',
            IDRef       => $Param{LockIDs},
        );
    }

    # current owner user ids
    if ( $Param{OwnerIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.user_id',
            IDRef       => $Param{OwnerIDs},
        );
    }

    # current responsible user ids
    if ( $Param{ResponsibleIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.responsible_user_id',
            IDRef       => $Param{ResponsibleIDs},
        );
    }


    # current sub queue ids
    if ( $Param{UseSubQueues} && $Param{QueueIDs} ) {
        my @SubQueueIDs;
        my %Queues = $Self->{QueueObject}->GetAllQueues();
        for my $QueueID ( @{ $Param{QueueIDs} } ) {
            my $Queue = $Self->{QueueObject}->QueueLookup( QueueID => $QueueID );
            for my $QueuesID ( sort keys %Queues ) {
                if ( $Queues{$QueuesID} =~ /^\Q$Queue\E::/i ) {
                    push @SubQueueIDs, $QueuesID;
                }
            }
        }
        push @{ $Param{QueueIDs} }, @SubQueueIDs;
    }

    # current queue ids
    if ( $Param{QueueIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.queue_id',
            IDRef       => $Param{QueueIDs},
        );
    }

    my @GroupIDs;

    # user groups
    if ( $Param{UserID} && $Param{UserID} != 1 ) {

        # get users groups
        @GroupIDs = $Self->{GroupObject}->GroupMemberList(
            UserID => $Param{UserID},
            Type   => $Param{Permission} || 'ro',
            Result => 'ID',
        );

        # return if we have no permissions
        return if !@GroupIDs;
    }

    # add group ids to sql string
    if (@GroupIDs) {
        $SQLExt .= " AND q.group_id IN (${\(join ', ' , sort {$a <=> $b} @GroupIDs)}) ";
    }

    # current priority lookup
    if ( $Param{Priorities} ) {
        for my $Priority ( @{ $Param{Priorities} } ) {

            # lookup priority id
            my $PriorityID = $Self->{PriorityObject}->PriorityLookup(
                Priority => $Priority,
            );
            return if !$PriorityID;
            push @{ $Param{PriorityIDs} }, $PriorityID;
        }
    }

    # priority ids
    if ( $Param{PriorityIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.ticket_priority_id',
            IDRef       => $Param{PriorityIDs},
        );
    }

    # service ids
    if ( $Param{ServiceIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.service_id',
            IDRef       => $Param{ServiceIDs},
        );
    }

    # sla ids
    if ( $Param{SLAIDs} ) {
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'st.sla_id',
            IDRef       => $Param{SLAIDs},
        );
    }

    if ( $Param{WatchUserIDs} ) {
        $SQLFrom .= ' LEFT JOIN ticket_watcher tw ON tw.ticket_id = st.id ';
        $SQLExt .= $Self->_ConditionGet(
            TableColumn => 'tw.user_id',
            IDRef       => [ $Param{WatchUserIDs}->[0] ],
        );
    }

    if ( $Param{CustomerName} ) {
        $SQLFrom .= ' LEFT JOIN customer_user cu ON cu.login = st.customer_user_id ';
    }

    # other ticket stuff
    my %FieldSQLMap = (
        TicketNumber      => 'st.tn',
        Title             => 'st.title',
        CustomerID        => 'st.customer_id',
        CustomerUserLogin => 'st.customer_user_id',
    );
    for my $Key ( sort keys %FieldSQLMap ) {

        # next if attribute is not used
        next if !defined $Param{$Key};

        # if it's no ref, put it to array ref
        if ( ref $Param{$Key} eq '' ) {
            $Param{$Key} = [ $Param{$Key} ];
        }

        # proccess array ref
        my $Used = 0;
        for my $Value ( @{ $Param{$Key} } ) {

            # next if no search attribute is given
            next if !$Value;

            # replace wild card search
            $Value =~ s/\*/%/gi;

            # check search attribute, we do not need to search for *
            next if $Value =~ /^\%{1,3}$/;

            if ( !$Used ) {
                $SQLExt .= ' AND (';
                $Used = 1;
            }
            else {
                $SQLExt .= ' OR ';
            }

            # add * to prefix/suffix on title search
            my %ConditionFocus;
            if ( $Param{ConditionInline} && $Key eq 'Title' ) {
                $ConditionFocus{Extended} = 1;
                if ( $Param{ContentSearchPrefix} ) {
                    $ConditionFocus{SearchPrefix} = $Param{ContentSearchPrefix};
                }
                if ( $Param{ContentSearchSuffix} ) {
                    $ConditionFocus{SearchSuffix} = $Param{ContentSearchSuffix};
                }
            }

            # use search condition extension
            $SQLExt .= $Self->{DBObject}->QueryCondition(
                Key   => $FieldSQLMap{$Key},
                Value => $Value,
                %ConditionFocus,
            );
        }
        if ($Used) {
            $SQLExt .= ')';
        }
    }


    # archive flag
    if ( $Self->{ConfigObject}->Get('Ticket::ArchiveSystem') ) {

        # if no flag is given, only search for not archived ticket
        if ( !$Param{ArchiveFlags} ) {
            $Param{ArchiveFlags} = ['n'];
        }

        # prepare search with archive flags, check arguments
        if ( ref $Param{ArchiveFlags} ne 'ARRAY' ) {
            $Self->{LogObject}->Log(
                Priority => 'error',
                Message  => "Invalid attribute ArchiveFlags '$Param{ArchiveFlags}'!",
            );
            return;
        }

        # prepare options
        my %Options;
        for my $Key ( @{ $Param{ArchiveFlags} } ) {
            $Options{$Key} = 1;
        }

        # search for archived
        if ( $Options{y} && !$Options{n} ) {
            $SQLExt .= ' AND archive_flag = 1';
        }

        # search for not archived
        elsif ( !$Options{y} && $Options{n} ) {
            $SQLExt .= ' AND archive_flag = 0';
        }
    }

    # database query for sort/order by option
    if ( $Result ne 'COUNT' ) {
        $SQLExt .= ' ORDER BY';
        for my $Count ( 0 .. $#SortByArray ) {
            if ( $Count > 0 ) {
                $SQLExt .= ',';
            }

            # regular sort
            # $SQLSelect .= ', ' . $SortOptions{ $SortByArray[$Count] };
            $SQLExt    .= ' ' . $SortOptions{ $SortByArray[$Count] };

            if ( $OrderByArray[$Count] eq 'Up' ) {
                $SQLExt .= ' ASC';
            }
            else {
                $SQLExt .= ' DESC';
            }
        }
    }

    # check cache
    my $CacheObject;
    if ( $Param{CacheTTL} ) {
        $CacheObject = Kernel::System::Cache->new( %{$Self} );
        my $CacheData = $CacheObject->Get(
            Type => 'TicketSearchSimple',
            Key  => $SQLSelect . $SQLFrom . $SQLExt . $Result . $Limit,
        );

        if ( defined $CacheData ) {
            if ( ref $CacheData eq 'HASH' ) {
                return %{$CacheData};
            }
            elsif ( ref $CacheData eq 'ARRAY' ) {
                return @{$CacheData};
            }
            elsif ( ref $CacheData eq '' ) {
                return $CacheData;
            }
            $Self->{LogObject}->Log(
                Priority => 'error',
                Message  => 'Invalid ref ' . ref($CacheData) . '!'
            );
            return;
        }
    }

    # database query
    my %Tickets;
    my $Count;

    return
        if !$Self->{DBObject}->Prepare( SQL => $SQLSelect . $SQLFrom . $SQLExt, Limit => $Limit );

    # return COUNT
    if ( $Result eq 'COUNT' ) {
        while ( my @Row = $Self->{DBObject}->FetchrowArray() ) {
            $Count = $Row[0];
        }
        if ($CacheObject) {
            $CacheObject->Set(
                Type  => 'TicketSearch',
                Key   => $SQLSelect . $SQLFrom . $SQLExt . $Result . $Limit,
                Value => $Count,
                TTL   => $Param{CacheTTL} || 60 * 4,
            );
        }
        return $Count;
    }
    else {
        while ( my @Row = $Self->{DBObject}->FetchrowArray() ) {
            for ( 0..$#FieldOptions ) {
                $Tickets{ $Row[0] }->{ $FieldOptions[$_]->{Key} } = $Row[$_];
            }

            $Tickets{ $Row[0] }{Age} = $Self->{TimeObject}->SystemTime() - $Tickets{ $Row[0] }{CreateTimeUnix};
            $Tickets{ $Row[0] }{OwnerFullName} = $Tickets{ $Row[0] }{OwnerName} . ' ' . $Tickets{ $Row[0] }{OwnerSurname};
            $Tickets{ $Row[0] }{CustomerFullName} = $Tickets{ $Row[0] }{CustomerName} . ' ' . $Tickets{ $Row[0] }{CustomerSurname}
                if $Tickets{ $Row[0] }{CustomerName} or $Tickets{ $Row[0] }{CustomerSurname};
            $Tickets{ $Row[0] }{Unread} = !$Tickets{ $Row[0] }{Readed};
        }

        # отдельным запросам приклеиваем динамические поля, если нужно
        if ( $Param{DynamicFields} ) {
            my $DynamicFields = $Self->_DynamicFieldsGet(
                IDList => [ keys %Tickets ],
                ObjectType => 'Ticket',
            );

            for my $TicketID ( keys %{ $DynamicFields } ) {
                for my $FieldName ( keys %{$DynamicFields->{$TicketID}} ) {
                    $Tickets{$TicketID}{$FieldName} = $DynamicFields->{$TicketID}->{$FieldName};
                }
            }
        }

        # return HASH
        if ( $Result eq 'HASH' ) {
            if ($CacheObject) {
                $CacheObject->Set(
                    Type  => 'TicketSearch',
                    Key   => $SQLSelect . $SQLFrom . $SQLExt . $Result . $Limit,
                    Value => \%Tickets,
                    TTL   => $Param{CacheTTL} || 60 * 4,
                );
            }
            return %Tickets;
        }

        # return ARRAY
        else {

            if ($CacheObject) {
                $CacheObject->Set(
                    Type  => 'TicketSearch',
                    Key   => $SQLSelect . $SQLFrom . $SQLExt . $Result . $Limit,
                    Value => [ values %Tickets ],
                    TTL   => $Param{CacheTTL} || 60 * 4,
                );
            }
            return values %Tickets;
        }
    }
}

=begin Internal:

=cut

=item _ConditionGet()

internal function to create an

    AND table.column IN (values)

condition string from an array.

    my $SQLPart = $TicketObject->_ConditionGet(
        TableColumn => 'table.column',
        IDRef       => $ArrayRef,
    );

=cut

sub _ConditionGet {
    my ( $Self, %Param ) = @_;

    # check needed stuff
    for my $Key (qw(TableColumn IDRef)) {
        if ( !$Param{$Key} ) {
            $Self->{LogObject}->Log(
                Priority => 'error',
                Message  => "Need $Key!",
            );
            return;
        }
    }

    # sort ids to cache the SQL query
    my @SortedIDs = sort { $a <=> $b } @{ $Param{IDRef} };

    # quote values
    for my $Value (@SortedIDs) {
        return if !defined $Self->{DBObject}->Quote( $Value, 'Integer' );
    }

    return " AND $Param{TableColumn} IN (" . ( join ',', @SortedIDs ) . ")";
}

=item _DynamicFieldsGet()

внутренняя функция для получения динамических полей из базы
по заданным идентификаторам тикетов

=cut

sub _DynamicFieldsGet {
    my ( $Self, %Param ) = @_;

    return unless defined $Param{IDList} && ref $Param[IDList] eq 'ARRAY' && scalar @{$Param{IDList}};
    return unless $Param{ObjectType};

    my %Result;
    my $Limit = $Param{Limit} || 1000;
    my $SQL = "SELECT dfv.object_id,  df.name, dfv.value_text, dfv.value_date, dfv.value_int
        FROM dynamic_field_value dfv
            JOIN dynamic_field df ON dfv.field_id = df.id
    ";

    my $SQLExt = " WHERE dfv.object_id IN ("
        . join( ', ', map { $Self->{DBObject}->Quote( $_, 'Integer' ) } @{ $Param{IDList} } )
        .") AND df.object_type = '"
        . $Self->{DBObject}->Quote( $Param{ObjectType} )
        . "'";


    $Self->{DBObject}->Prepare( SQL => $SQL . $SQLExt, Limit => $Limit );

    # TODO: будут проблемы с мультиселектом
    while ( my @Row = $Self->{DBObject}->FetchrowArray() ) {
        $Result{ $Row[0] }->{ 'DynamicField_' . $Row[1] } = $Row[2] || $Row[3] || $Row[4] || '';
    }

    return \%Result;
}

1;

=end Internal:

=back

=head1 TERMS AND CONDITIONS

This software is part of the OTRS project (L<http://otrs.org/>).

This software comes with ABSOLUTELY NO WARRANTY. For details, see
the enclosed file COPYING for license information (AGPL). If you
did not receive this file, see L<http://www.gnu.org/licenses/agpl.txt>.

=head1 VERSION

$Revision: 1.21 $ $Date: 2013/01/10 15:20:08 $

=cut
