# --
# Kernel/Output/HTML/DashboardTicketSlickGrid.pm
# Copyright (C) 2013 Меркушов Виктор merkushov.victor@gmail.com
# --
# This software comes with ABSOLUTELY NO WARRANTY. For details, see
# the enclosed file COPYING for license information (AGPL). If you
# did not receive this file, see http://www.gnu.org/licenses/agpl.txt.
# --

package Kernel::Output::HTML::DashboardTicketSlickGrid;

use strict;
use warnings;

use vars qw($VERSION);
$VERSION = qw($Revision: 1.00 $) [1];

sub new {
    my ( $Type, %Param ) = @_;

    # allocate new hash for object
    my $Self = {%Param};
    bless( $Self, $Type );

    # get needed objects
    for (
        qw(Config Name ConfigObject LogObject DBObject LayoutObject ParamObject TicketObject UserID)
        )
    {
        die "Got no $_!" if ( !$Self->{$_} );
    }

    # get current filter
    my $Name = $Self->{ParamObject}->GetParam( Param => 'Name' ) || '';
    my $PreferencesFilterKey = 'UserDashboard' . $Self->{Name} . '-TicketGenericFilter';

    if ( $Self->{Name} eq $Name ) {
        $Self->{Filter} = $Self->{ParamObject}->GetParam( Param => 'Filter' ) || '';
    }

    # remember filter
    if ( $Self->{Filter} ) {

        # update session
        $Self->{SessionObject}->UpdateSessionID(
            SessionID => $Self->{SessionID},
            Key       => $PreferencesFilterKey,
            Value     => $Self->{Filter},
        );

        # update preferences
        $Self->{UserObject}->SetPreferences(
            UserID => $Self->{UserID},
            Key    => $PreferencesFilterKey,
            Value  => $Self->{Filter},
        );

    } else {
        $Self->{Filter} = $Self->{$PreferencesFilterKey} || $Self->{Config}->{Filter} || 'All';
    }

    # Ключи
    $Self->{FieldKey}   = 'UserDashboard' . $Self->{Name} . '-Field';
    $Self->{FilterKey}  = 'UserDashboard' . $Self->{Name} . '-Filter';

    my @FieldShown = ('TicketNumber', 'Title');
    if ( $Self->{LayoutObject}->{ $Self->{FieldKey} } ) {
        @FieldShown = split( ',', $Self->{LayoutObject}->{ $Self->{FieldKey} } );
    }
    $Self->{FieldShown} = \@FieldShown;

    my @FilterShown = split( ',', $Self->{LayoutObject}->{ $Self->{FilterKey} } );
    $Self->{FilterShown} = \@FilterShown;

    return $Self;
}

sub Preferences {
    my ( $Self, %Param ) = @_;

    my @Params = (
        {
            Desc  => 'Filters available in the widget',
            Name  => $Self->{FilterKey},
            Block => 'Checkbox',
            Data => $Self->_FilterConfig(),
            SelectedID  => $Self->{FilterShown},
            Translation => 1,
        },
        {
            Desc  => 'Fields available in the widget',
            Name  => $Self->{FieldKey},
            Block => 'Checkbox',
            Data => $Self->_FieldConfig( FieldType => 'Common' ),
            SelectedID  => $Self->{FieldShown},
            Translation => 1,
        },
    );

    if ( $Self->{Config}->{DynamicFields} ) {
        push @Params, {
            Desc  => 'Additional Dynamic Fields available in the widget',
            Name  => $Self->{FieldKey},
            Block => 'Checkbox',
            Data => $Self->_FieldConfig( FieldType => 'Dynamic' ),
            SelectedID  => $Self->{FieldShown},
            Translation => 1,
        };
    }

    return @Params;
}

sub _FilterConfig {
    my ( $Self, %Param ) = @_;

    my @Filters = (
        { Key => 'MyLocked',    Value => 'My locked tickets' },
        { Key => 'Responsible', Value => 'My responsibilities' },
        { Key => 'Watcher',     Value => 'My watched tickets' },
        { Key => 'MyQueues',    Value => 'Tickets in My Queues' },
        { Key => 'All',         Value => 'All tickets', },
    );

    my @NewFilters;
    for my $Filter ( @Filters ) {
        next if $Filter->{Key} eq 'Responsible' && !$Self->{ConfigObject}->Get('Ticket::Responsible');
        next if $Filter->{Key} eq 'Watcher' && !$Self->{ConfigObject}->Get('Ticket::Watcher');
        push @NewFilters, $Filter;
    }

    return \@NewFilters;
}

sub _FieldConfig {
    my ( $Self, %Param ) = @_;

    my @CommonFieldsConfig = (
        { Key => 'Priority',        Value => 'Priority',    Width => 70 },
        { Key => 'Unread',          Value => 'Unread',      Width => 70, Formatter => 'FormatterUnread' },

        { Key => 'TicketNumber',    Value => 'Ticket#',     Width => 50 },
        { Key => 'Title',           Value => 'Title',       Width => 300, MinWidth => 200 },
        { Key => 'Queue',           Value => 'Queue',       Width => 300, MinWidth => 200 },
        { Key => 'State',           Value => 'State',       Width => 100 },
        { Key => 'CustomerID',      Value => 'CustomerID',  Width => 120 },
        { Key => 'CustomerFullName',  Value => 'Customer',  Width => 120 },
        { Key => 'OwnerFullName',   Value => 'Owner',       Width => 150 },
        { Key => 'Type',            Value => 'Type',        Width => 100 },
        { Key => 'SLA',             Value => 'SLA.',        Width => 100 },
        { Key => 'Service',         Value => 'Service',     Width => 150 },
        { Key => 'Responsible',     Value => 'Responsible', Width => 100 },
        { Key => 'Age',             Value => 'Age',         Width => 100, Formatter => 'FormatterAge' },
        { Key => 'Created',         Value => 'Created',     Width => 100, Formatter => 'FormatterDate' },
        { Key => 'Changed',         Value => 'Changed',     Width => 100, Formatter => 'FormatterDate' },
    );

    my %Show = map { $_ => 1 } @{ $Param{ShowOnly} };

    my @Fields;
    if ( !defined $Param{'FieldType'} or $Param{'FieldType'} eq 'Common' ) {
        for my $Field ( @CommonFieldsConfig ) {
            next if $Field->{Key} eq 'Responsible' && !$Self->{ConfigObject}->Get('Ticket::Responsible');
            next if $Field->{Key} eq 'Watcher' && !$Self->{ConfigObject}->Get('Ticket::Watcher');
            next if scalar @{$Param{ShowOnly}} && !$Show{ $Field->{Key} };

            push @Fields, $Field;
        }
    }

    if ( !defined $Param{'FieldType'} or $Param{'FieldType'} eq 'Dynamic' ) {
        my $DynamicFieldsConfig = $Self->{Config}->{DynamicFields};
        for my $FieldKey ( keys %{$DynamicFieldsConfig} ) {
            my $Field = {
                Key     => $FieldKey,
                Value   => $DynamicFieldsConfig->{$FieldKey},
                Width   => 100
            };
            next if scalar @{$Param{ShowOnly}} && !$Show{ $Field->{Key} };

            push @Fields, $Field;
        }
    }

    return \@Fields;
}

sub Config {
    my ( $Self, %Param ) = @_;

    # check if frontend module of link is used
    if ( $Self->{Config}->{Link} && $Self->{Config}->{Link} =~ /Action=(.+?)([&;].+?|)$/ ) {
        my $Action = $1;
        if ( !$Self->{ConfigObject}->Get('Frontend::Module')->{$Action} ) {
            $Self->{Config}->{Link} = '';
        }
    }

    return (
        %{ $Self->{Config} },

        # remember, do not allow to use page cache
        # (it's not working because of internal filter)
        CacheTTL => undef,
        CacheKey => undef,
    );
}

sub Run {
    my ( $Self, %Param ) = @_;

    my $CacheKey = $Self->{Name} . '-' . $Self->{UserID};

    # get all search base attributes
    my %TicketSearch;
    my %DynamicFieldsParameters;
    my @Params = split /;/, $Self->{Config}->{Attributes};
    for my $String (@Params) {
        next if !$String;
        my ( $Key, $Value ) = split /=/, $String;

        # push ARRAYREF attributes directly in an ARRAYREF
        if (
            $Key
            =~ /^(StateType|StateTypeIDs|Queues|QueueIDs|Types|TypeIDs|States|StateIDs|Priorities|PriorityIDs|Services|ServiceIDs|SLAs|SLAIDs|Locks|LockIDs|OwnerIDs|ResponsibleIDs|WatchUserIDs|ArchiveFlags)$/
            )
        {
            if ( $Value =~ m/,/ ) {
                push @{ $TicketSearch{$Key} }, split( /,/, $Value );
            } else {
                push @{ $TicketSearch{$Key} }, $Value;
            }
        }
        elsif ( !defined $TicketSearch{$Key} ) {
            $TicketSearch{$Key} = $Value;
        }
        elsif ( !ref $TicketSearch{$Key} ) {
            my $ValueTmp = $TicketSearch{$Key};
            $TicketSearch{$Key} = [$ValueTmp];
            push @{ $TicketSearch{$Key} }, $Value;
        }
        else {
            push @{ $TicketSearch{$Key} }, $Value;
        }
    }

    %TicketSearch = (
        %TicketSearch,
        Permission => $Self->{Config}->{Permission} || 'ro',
        UserID => $Self->{UserID},
    );

    # CustomerInformationCenter shows data per CustomerID
    if ( $Param{CustomerID} ) {
        $TicketSearch{CustomerID} = $Param{CustomerID};
        $CacheKey .= '-' . $Param{CustomerID};
    }

    # define filter attributes
    my @MyQueues = $Self->{QueueObject}->GetAllCustomQueues(
        UserID => $Self->{UserID},
    );
    if ( !@MyQueues ) {
        @MyQueues = (999_999);
    }
    my %TicketSearchSummary = (
        MyLocked => {
            OwnerIDs => [ $Self->{UserID}, ],
            LockIDs  => [ 2 ],
        },
        Watcher => {
            WatchUserIDs => [ $Self->{UserID}, ],
            LockIDs      => undef,
        },
        Responsible => {
            ResponsibleIDs => [ $Self->{UserID}, ],
            LockIDs        => undef,
        },
        MyQueues => {
            QueueIDs => \@MyQueues,
            LockIDs  => undef,
        },
        All => {
            OwnerIDs => undef,
            LockIDs  => undef,
        },
    );

    if ( defined $TicketSearch{QueueIDs} ) {
        delete $TicketSearchSummary{MyQueues};
    }

    my $CacheUsed = 1;

    # find and show ticket list
    my @Tickets = $Self->{TicketObject}->TicketSearchSimple(
        Result          => 'ARRAY',
        DynamicFields   => 1,
        CustomerName    => 1,
        %TicketSearch,
        %{ $TicketSearchSummary{ $Self->{Filter} } },
    );

    # check cache
    my $Summary = $Self->{CacheObject}->Get(
        Type => 'Dashboard',
        Key  => $CacheKey . '-Summary',
    );

    unless ( $Summary ) {
        for my $Type ( sort keys %TicketSearchSummary ) {
            next if !$TicketSearchSummary{$Type};

            $Summary->{$Type} = $Self->{TicketObject}->TicketSearchSimple(
                Result => 'COUNT',
                %TicketSearch,
                %{ $TicketSearchSummary{$Type} },
            );
        }
    }

    # set cache
    if ( !$CacheUsed && $Self->{Config}->{CacheTTLLocal} ) {
        $Self->{CacheObject}->Set(
            Type  => 'Dashboard',
            Key   => $CacheKey . '-Summary',
            Value => $Summary,
            TTL   => $Self->{Config}->{CacheTTLLocal} * 60,
        );
    }

    # set css class
    $Summary->{ $Self->{Filter} . '::Selected' } = 'Selected';

    if ( scalar @{$Self->{FilterShown}} ) {
        # get filter ticket counts
        $Self->{LayoutObject}->Block(
            Name => 'ContentLargeTicketGenericFilter',
            Data => {
                %Param,
                %{ $Self->{Config} },
                Name => $Self->{Name},
                %{$Summary},
            },
        );
    }

    for my $ShowFilter ( @{$Self->{FilterShown}} ) {

        my $FilterName;
        if ( $ShowFilter eq 'All' ) {
            $FilterName = 'ContentLargeTicketGenericFilterAll';
        } elsif ( $ShowFilter eq 'MyLocked' ) {
            $FilterName = 'ContentLargeTicketGenericFilterMyLocked';
        } elsif ( $Self->{ConfigObject}->Get('Ticket::Watcher') and $ShowFilter eq 'Watcher' ) {
            $FilterName = 'ContentLargeTicketGenericFilterWatcher';
        } elsif ( $Self->{ConfigObject}->Get('Ticket::Responsible') and $ShowFilter eq 'Responsible' ) {
            $FilterName = 'ContentLargeTicketGenericFilterResponsible';
        } elsif ( $TicketSearchSummary{MyQueues} and $ShowFilter eq 'MyQueues' ) {
            $FilterName = 'ContentLargeTicketGenericFilterMyQueues';
        }

        $Self->{LayoutObject}->Block(
            Name => $FilterName,
            Data => {
                %Param,
                %{ $Self->{Config} },
                Name => $Self->{Name},
                %{$Summary},
            },
        );
    }

    my @GridColumns;
    my @GridData;
    my @Fields = @{ $Self->_FieldConfig( ShowOnly => $Self->{FieldShown} ) };

    # show block title
    if ( @Fields ) {
        my $BlockTitle;
        for my $Field ( @Fields ) {

            push @GridColumns, {
                id          => $Field->{Key},
                name        => $Field->{NoTitle} ? '' : $Self->{LayoutObject}->{LanguageObject}->Get( $Field->{Value} ),
                field       => $Field->{Key},
                sortable    => 'true',
                ( $Field->{Width}       ? (width => $Field->{Width}) : () ),
                ( $Field->{MinWidth}    ? (minWidth => $Field->{MinWidth}) : () ),
                ( $Field->{MaxWidth}    ? (maxWidth => $Field->{MaxWidth}) : () ),
                ( $Field->{Formatter}   ? (formatter => $Field->{Formatter}) : () ),
            };

        }
    }

    # show tickets
    my $Count = 0;
    for my $Ticket ( @Tickets ) {
        $Count++;
        $GridData[$Count-1]->{'id'} = $Ticket->{TicketID};

        for my $Field ( @Fields ) {
            my $FieldValue = $Ticket->{$Field->{Key}};
            if ( $Field->{Key} eq 'State' or $Field->{Key} eq 'Priority' ) {
                $FieldValue = $Self->{LayoutObject}->{LanguageObject}->Get( $FieldValue );
            }
            $GridData[$Count-1]->{$Field->{Key}} = $FieldValue || '';
        }

    }

    my $Content = $Self->{LayoutObject}->Output(
        TemplateFile => 'AgentDashboardTicketSlickGrid',
        Data         => {
            %{ $Self->{Config} },
            %{$Summary},
            Name                => $Self->{Name},
            AvailableFilters    => join( ', ', map { "'$_'" } @{$Self->{FilterShown}} ),
            CurrentFilter       => $Self->{Filter} || 'All',
            GridColumns         => $Self->{LayoutObject}->JSONEncode( Data => \@GridColumns ),
            GridData            => $Self->{LayoutObject}->JSONEncode( Data => \@GridData ),
        },
        KeepScriptTags => $Param{AJAX},
    );

    return $Content;
}

1;
